#ifndef GLOBAL_H
#define GLOBAL_H

#include <array>
#include <cmath>
#include <string>

constexpr int kboard_size = 9;
constexpr int kboard_square_size = kboard_size * kboard_size;
constexpr int kexpanded_board_size = kboard_size + 4;
constexpr int karray_board_size = kexpanded_board_size * kexpanded_board_size;
constexpr int kboard_start_index = 2 * kexpanded_board_size;
constexpr int kboard_end_index = karray_board_size - 2 * kexpanded_board_size;


constexpr int kthread_size = 7;
constexpr long long int knode_size = 2000000;

constexpr int kdepth_one_node_size = kboard_square_size  + kthread_size;

constexpr int ktrial_period = 300;

constexpr int kselect_width = 30;

constexpr int max_numerator = 20;
constexpr int max_denominator = 20;
constexpr int knode_max_width = kboard_square_size + 1; // 1 : pass

constexpr int kkomi = 6;
constexpr double kreal_komi = 6.5;

constexpr int krandom_table_size = 10;

enum {
    Black = 0,
    White,
    Empty,
    Wall,
};

void printPosition(int pos);
void doCalucUCBBias();
std::string getCharacterPosition(int pos);

inline int getEast(int pos) { return pos + 1; }
inline int getWest(int pos) { return pos - 1; }
inline int getNorth(int pos) { return pos - kexpanded_board_size; }
inline int getSouth(int pos) { return pos + kexpanded_board_size; }

inline int getSouthWest(int pos) { return pos + kexpanded_board_size - 1; }
inline int getNorthWest(int pos) { return pos - kexpanded_board_size - 1; }
inline int getSouthEast(int pos) { return pos + kexpanded_board_size + 1; }
inline int getNorthEast(int pos) { return pos - kexpanded_board_size + 1; }

inline int getManhattanDistance(int from, int to)
{
    int from_x = from % kexpanded_board_size;
    int from_y = from / kexpanded_board_size;
    int to_x = to % kexpanded_board_size;
    int to_y = to / kexpanded_board_size;

    return (abs(from_x - to_x) + abs(from_y - to_y));
}

const int board9x9_index[81] = {
    28, 29, 30, 31, 32, 33, 34, 35, 36,
    41, 42, 43, 44, 45, 46, 47, 48, 49,
    54, 55, 56, 57, 58, 59, 60, 61, 62,
    67, 68, 69, 70, 71, 72, 73, 74, 75,
    80, 81, 82, 83, 84, 85, 86, 87, 88,
    93, 94, 95, 96, 97, 98, 99, 100, 101,
    106, 107, 108, 109, 110, 111, 112, 113, 114,
    119, 120, 121, 122, 123, 124, 125, 126, 127,
    132, 133, 134, 135, 136, 137, 138, 139, 140 };

const int board9x9_upper_side_index[9] = {
    28, 29, 30, 31, 32, 33, 34, 35, 36
};

const int board9x9_under_side_index[9] = {
    132, 133, 134, 135, 136, 137, 138, 139, 140
};

const int board9x9_left_side_index[9] = {
    28, 41, 54, 67, 80, 93, 106, 119, 132
};

const int board9x9_right_side_index[9] = {
    36, 49, 62, 75, 88, 101, 114, 127, 140
};

const int board9x9_second_upper_side_index[9] = {
    41, 42, 43, 44, 45, 46, 47, 48, 49,
};

const int board9x9_second_under_side_index[9] = {
    119, 120, 121, 122, 123, 124, 125, 126, 127
};

const int board9x9_second_left_side_index[9] = {
    29, 42, 55, 68, 81, 94, 107, 120, 133
};

const int board9x9_second_right_side_index[9] = {
    35, 48, 61, 74, 87, 100, 113, 126, 139
};

const int board9x9_edge_index[32] = {
    28, 29, 30, 31, 32, 33, 34, 35, 36,
    41, 49, 54, 62, 67, 75, 80, 88, 93,
    101, 106, 114, 119, 127,
    132, 133, 134, 135, 136, 137, 138, 139, 140};

const int board9x9_result_index[9] = { 42, 45, 48, 81, 84, 87, 120, 123, 126 };
const bool kis_use_board9x9_result_index = true;


#endif // GLOBAL_H
