#ifndef GO_BOARD_H
#define GO_BOARD_H

#include <array>
#include <vector>
#include <string>
#include <bitset>
#include "global.h"

class GoBoard
{
public:
    GoBoard();
    void initializeGoBoard();
    void printGoBoard() const;
    void setStone(int pos, int color);
    void printGoStrings();
    bool isLegalMove(int pos, int color) const;
    void readSGF(std::string file_name);
    bool isEye(int pos, int color);
    int getPrePosition() const { return pre_position_; }
    int getSecondPrePosition() const { return second_pre_position_; }
    int getThirdPrePosition() const { return thrid_pre_position_; }

    int getPointColor(int pos) const { return go_board_[pos]; }
    void initializeKoPoint() { ko_position_ = 0; }
    int getLibertyCount(int pos) const {
        // for eye
        if (go_board_[pos] == Empty || go_board_[pos] == Wall) {
            return 5;
        }
        return go_string_liberty_count_[go_string_board_[pos]];
    }
    int getBitPattern(int pos) { return pattern_go_board_[pos]; }

    int getResultChineseRule(int color);
    double getResultScoreChineseRule();
    void printPatternGoBoard() const ;
    void printPattern3x3GoBoard(int pos) const;
    void setPatternGoBoard();
    void setPointPatternGoBoard(int pos);
    std::array<int, karray_board_size> getPatternGoBoard() const { return pattern_go_board_; }
    bool isNick(int pos, int color);
    int getStoneCount() { return stone_count_; }

private:

    void clearGoBoard();
    void clearGoStrings();

    int ko_position_;
    int pre_position_;
    int second_pre_position_;
    int thrid_pre_position_;

    std::array<int, karray_board_size> go_board_;
    std::array<int, karray_board_size> pattern_go_board_;
    std::vector<int> go_string_liberty_count_;
    // std::array<int, 3 *  karray_board_size> go_string_liberty_count_;
    std::array<int,  karray_board_size> go_string_board_;
    int go_string_index_[2]; // 0 : black, 1 : which
    int stone_count_;

    void setBitPattern(int pos, int color, int bit_pos);
public:
    void printProbability(std::array<double,  karray_board_size> board);
    void printProbabilityCount(std::array<double,  karray_board_size> board);
    std::array<int,  karray_board_size> getGoBoard() const { return go_board_; }
};

#endif // GO_BOARD_H
