#ifndef GTP_H
#define GTP_H

#include <string>
#include <unordered_map>
#include "go_board.h"
#include "search.h"
#include "move_frequency.h"
#include "make_pattern.h"

class GTP
{
public:
    GTP();
    void loop();
private:
    enum Status {
        OK,
        FAILURE,
        QUIT,
    };

    std::string ai_name_;
    GoBoard go_board_;
    MoveFrequency move_frequency_;

    int turn_count_;
    int turn_color_;
    Search search_;

    typedef Status (GTP::*func)(const std::string &);
    std::unordered_map<std::string, func> handler_;

    Status success(const std::string &);
    Status failure(const std::string &);
    Status name(const std::string &);
    Status quit(const std::string &);
    Status genmove(const std::string &);
    Status clearBoard(const std::string &);
    Status boardSize(const std::string &);
    Status listCommands(const std::string &);
    Status play(const std::string &);
    Status goguiAnalyzeCommands(const std::string &);
    Status ProbabilityAnalyst(const std::string &);
    Status ProbabilityPattern3x3(const std::string &);
    Status ProbabilityCountPattern(const std::string &);
    Status ProbabilityPattern5x5(const std::string &);
    Status ProbabilityPattern9x9(const std::string &);
    Status doTestSearch(const std::string &);
    Status finalScore(const std::string &);
    Status ProbabilityPattern3x3WithPosition(const std::string &);
    Status ProbabilityPattern(const std::string &);
};

#endif // GTP_H
