#include <dirent.h>
#include <sstream>
#include <fstream>
#include <iostream>
#include <array>
#include <algorithm>
#include <cstdlib>
#include <cmath>
#include "go_board.h"
#include "global.h"
#include "move_frequency.h"


MoveFrequency::MoveFrequency()
{
}

void MoveFrequency::doPreparation()
{
    DIR *dir;
    int color = Black;
    char ch, x, y;
    int num_x, num_y;

    struct dirent *dp;
    std::string path = "/home/move_frequency";
    if ((dir = opendir(path.c_str())) == NULL) {
        std::cerr << "not found directory" << std::endl;
        exit(1);
    }

    // initialization of frequency_board
    for (int i = 0; i < 2; ++i) {
        for (int j = 0; j <  karray_board_size; ++j) {
            for (int k = 0; k <  karray_board_size; ++k) {
                frequency_board_[i][j][k].fill(0.0);
            }
        }
    }

    while ((dp = readdir(dir)) != NULL) {
        if (dp->d_name[0] == '.') {
            continue;
        }
        std::string file_name(dp->d_name);
        if (file_name.find(".sgf") == std::string::npos) {
            continue;
        }

        file_name = path + "/" + file_name;
        std::fstream fin(file_name.c_str());

        GoBoard go_board;
        int pre_pos = -1;
        while (fin) {
            fin >> ch;
            if (ch == ';') {
                fin >> ch;
                if (ch == 'B' || ch == 'W') {
                    fin >> ch;
                    if (ch == '[') {
                        fin >> x;
                        if (x != ']') {
                            fin >> y>> ch;
                            if (ch == ']') {
                                num_x = x - 'a' + 2;
                                num_y = y - 'a' + 2;
                                num_y *= kexpanded_board_size;
                                int pos = num_x + num_y;

                                std::array<int,  karray_board_size> current_board = go_board.getGoBoard();
                                if (pre_pos != -1 && getManhattanDistance(pre_pos, pos) <= 5) {
                                    for (int i = 1; i <= kboard_size; ++i) {
                                        for (int j = 1; j <= kboard_size; ++j) {
                                            int index = i * kexpanded_board_size + j;
                                            if (index == pos) {
                                                continue;
                                            }

                                            if (current_board[index] == Black) {
                                                ++frequency_board_[color][pos][index][0];
                                            }

                                            else if (current_board[index] == White) {
                                                ++frequency_board_[color][pos][index][1];
                                            }
                                            else if (current_board[index] == Empty) {
                                                ++frequency_board_[color][pos][index][2];
                                            }
                                        }
                                    }
                                }
                                pre_pos = pos;
                                go_board.setStone(pos, color);
                            }
                        }
                        color = (color == Black ? White : Black);
                    }
                }
            }
            else if (ch == 'S') {
                fin >> ch;
                if (ch == 'Z') {
                    fin >> ch >> num_x;
                    if (num_x != kboard_size) {
                        break;
                    }
                }
            }
        }

        fin.close();
    }

    closedir(dir);
}

void MoveFrequency::printAroundFurequency(int pos, int color)
{
    for (int i = 1; i <= kboard_size; ++i) {
        for (int j = 1; j <= kboard_size; ++j) {
            int index = i * kexpanded_board_size + j;
            std::cerr << "[" << frequency_board_[color][pos][index][0] << ","
                      << frequency_board_[color][pos][index][1] << ","
                      << frequency_board_[color][pos][index][2] << "]";
        }
        std::cout << std::endl;
    }
}

int MoveFrequency::getManhattanDistance(int from, int to)
{
    int from_x = from % kexpanded_board_size;
    int from_y = from / kexpanded_board_size;
    int to_x = to % kexpanded_board_size;
    int to_y = to / kexpanded_board_size;

    return (abs(from_x - to_x) + abs(from_y - to_y));
}

std::array<double,  karray_board_size> MoveFrequency::getFrequencyBoard(std::array<int,  karray_board_size> go_board, int color)
{
    std::array<double,  karray_board_size> frequency_board, ranked_board, tmp_board;
    frequency_board.fill(0.0);
    double sum_frequency = 0.0;
    int empty_coutn;
    for (int i = kboard_size; i <  karray_board_size; ++i) {
        if (go_board[i] != Empty) {
            continue;
        }
        if (!(i % kexpanded_board_size)) {
            std::cerr << std::endl;
        }

        empty_coutn = 0;
        int max_point = 0, min_point = 0, actual_point = 0;
        for (int j = kboard_size; j <  karray_board_size; ++j) {
            if (go_board[j] == Wall || j == i) {
                continue;
            }
            int i_x = i % kexpanded_board_size;
            int i_y = i / kexpanded_board_size;
            int j_x = j % kexpanded_board_size;
            int j_y = j / kexpanded_board_size;
            if ((abs(i_x - j_x) + abs(i_y - j_y)) >= 5) {
                continue;
            }

            if (go_board[j] == Empty) {
                ++empty_coutn;
            }
            if (empty_coutn > 5) {
                continue;
            }

            max_point += *std::max_element(frequency_board_[color][i][j].begin(), frequency_board_[color][i][j].end());
            min_point += *std::min_element(frequency_board_[color][i][j].begin(), frequency_board_[color][i][j].end());
            if (go_board[j] == Black) {
                actual_point += frequency_board_[color][i][j][0];
            }
            else if (go_board[j] == White) {
                actual_point += frequency_board_[color][i][j][1];
            }
            else if (go_board[j] == Empty) {
                actual_point += frequency_board_[color][i][j][2];
            }
        }
        if (max_point == 0) {
            frequency_board[i] = 0.0;
        }
        else {
            frequency_board[i] = (double)(actual_point - min_point) / (double)max_point;
            std::cerr << "[" << max_point << ", " << min_point << ", " << actual_point << ", " << frequency_board[i] << "] ";
            sum_frequency += frequency_board[i];
        }
    }
    std::cerr << sum_frequency <<  std::endl << std::endl;

    tmp_board = frequency_board;
    ranked_board.fill(0);


    for (int i = kexpanded_board_size; i <  karray_board_size - kexpanded_board_size; ++i) {
        ranked_board[i] = frequency_board[i];//  / sum_frequency;
    }

    return ranked_board;
    // return frequency_board;
}






