#include <iostream>
#include <array>
#include <chrono>
#include <random>
#include <cmath>
#include <algorithm>
#include <memory>
#include <queue>
#include <atomic>
#include <cstdlib>
#include "ucb_node.h"
#include "uct_thread.h"
#include "go_board.h"
#include "make_pattern.h"
#include "not_read.h"

extern MakePattern make_pattern;
NotRead not_read;

UCTThread::UCTThread()
{

    // initialization of random
    std::random_device rnd;

    // vector for initializatioon
    std::vector<uint_least32_t> v(10);
    generate(v.begin(), v.end(), ref(rnd));

    // generate random engine
    std::seed_seq seed(v.begin(), v.end());
    engine_ = std::mt19937(seed);

    random_table_distribution_ = std::uniform_int_distribution<long long int>(0, krandom_table_size - 1);
    pibot_distribution_ = std::uniform_int_distribution<int>(0, 80);

    normal_distributuion_ = std::normal_distribution<>(0, 40);
}

void UCTThread::operator ()(const GoBoard &go_board,const int color, std::array<UCBNode, knode_size> *nodes,
                            int depth_one_node_end, int thread_number, int turn)
{
    nodes_ = nodes;
    thread_number_ = thread_number;
    use_node_max_index_ = depth_one_node_end;
    depth_one_node_end_ = depth_one_node_end;
    turn_ = turn;

    long long int node_index;
    long long int pre_node_index;
    int max_depth = 0;

    // std::cerr << "thread start " << thread_number << std::endl;

    auto to = std::chrono::system_clock::now() + std::chrono::milliseconds(950);
    int n = 0;
    while (std::chrono::system_clock::now() < to) {
        go_board_ = go_board;
        color_ = color;
        depth_ = 0;
        node_index = 0;
        int first_move_index = 0;
        int first_move_count = 0;

        for (;;) {
            pre_node_index = node_index;
            node_index = (*nodes_)[node_index].child_index_;


            node_index = getMaxUCBValueIndex(pre_node_index, node_index, knode_max_width);
            if (node_index == -1) {
                if (pre_node_index == 1) {
                    return;
                }
                else {
                    node_index = pre_node_index;
                    --depth_;
                    break;
                }
            }


            if (depth_ == 0) {
                first_move_index = node_index;
                first_move_count = (*nodes_)[node_index].trial_count_;
            }

            go_board_.setStone((*nodes_)[node_index].pos_, color_);
            color_ = (color_ == Black ? White : Black);
            ++depth_;


            if ((*nodes_)[node_index].child_index_ == 0) {
                break;
            }
        }

        if (use_node_max_index_ < (knode_size - karray_board_size)) {
            if ((*nodes_)[node_index].child_index_ == 0) {
                makeChildNode(node_index, color_);
            }
        }

        if (max_depth < depth_) {
            max_depth = depth_;
        }

        doOnePlayOut();

        double score = go_board_.getResultScoreChineseRule();
        int result;

        if (color == White) {
            result = (score < 0.0 ? 1 : 0);
            score *= -1;
        }
        else {
            result = (score > 0.0 ? 1 : 0);
        }


        (*nodes_)[first_move_index].score_ += score;

        if ((depth_ & 1) == 0) {
            result = (result == 0 ? 1 : 0);
        }
        updateNode(node_index, result);

        ++n;
    }
#ifdef EBUG
    std::cerr << "n"  << n << std::endl;
    std::cerr << "max depth " << max_depth << std::endl;
    std::cerr << "use " << use_node_max_index_ << std::endl;
    std::cerr << "node 0 : " << (*nodes_)[0].trial_count_ << std::endl;
    std::cerr << "thread end " << thread_number << std::endl;
#endif
}

long long int UCTThread::getMaxUCBValueIndex(long long int parent_index, long long int start_index, long long int range)
{
    if (parent_index != (*nodes_)[start_index].parent_index_) {
        (*nodes_)[parent_index].child_index_ = 0;
        return -1;
    }
    if ((*nodes_)[parent_index].depth_ + 1 != (*nodes_)[start_index].depth_) {
        (*nodes_)[parent_index].child_index_ = 0;
        return -1;
    }

    if (range == 0) {
        range = 1;
    }
    else if (range > kboard_square_size) {
        range = kboard_square_size + 1;
    }

    int sub_range = 0;
    int max_ucb_value_index = -1;
    double tmp_ucb_value;
    double max_ucb_value = -1.0;
    int parent_trial_count = (*nodes_)[parent_index].trial_count_;

    if ((*nodes_)[parent_index].depth_ == 0) {
        for (long long int i = 1; i < depth_one_node_end_; ++i) {
            if ((*nodes_)[i].is_legal_move_  == true) {

                if ((*nodes_)[i].trial_count_ == 0) {
                    if (not_read.isNotRead(go_board_, (*nodes_)[i].pos_, color_,  turn_) == false
                            && go_board_.isLegalMove((*nodes_)[i].pos_, color_) == true) {
                        return i;
                    }
                    else {
                        (*nodes_)[i].is_legal_move_ = false;
                    }
                }
                else if ((*nodes_)[i].trial_count_ > 0) {
                    if ((tmp_ucb_value = (*nodes_)[i].getUCBValue(parent_trial_count)) > max_ucb_value) {
                        max_ucb_value = tmp_ucb_value;
                        max_ucb_value_index = i;
                    }
                }
                else if ((tmp_ucb_value = (*nodes_)[i].getUCBValue(parent_trial_count)) > max_ucb_value) {
                    if ((go_board_.isEye((*nodes_)[i].pos_, color_) == false || go_board_.isNick((*nodes_)[i].pos_, color_) == true)
                            && go_board_.isLegalMove((*nodes_)[i].pos_, color_) == true) {
                        max_ucb_value = tmp_ucb_value;
                        max_ucb_value_index = i;
                    }
                    else {
                        (*nodes_)[i].is_legal_move_ = false;
                    }
                }
            }
        }
    }
    else {
        if (parent_trial_count < ktrial_period) {
            for (long long int i = start_index; i < range + start_index; ++i) {
                if ((*nodes_)[i].is_legal_move_ == true) {
                    if ((*nodes_)[i].trial_count_ == 0) {
                        if (not_read.isNotRead(go_board_, (*nodes_)[i].pos_, color_, turn_) == false
                                && go_board_.isLegalMove((*nodes_)[i].pos_, color_) == true) {
                            return i;
                        }
                        else {
                            (*nodes_)[i].is_legal_move_ = false;
                        }
                    }
                    else if ((*nodes_)[i].trial_count_ > 0) {
                        if ((tmp_ucb_value = (*nodes_)[i].getUCBValue(parent_trial_count)) > max_ucb_value) {
                            max_ucb_value = tmp_ucb_value;
                            max_ucb_value_index = i;
                        }
                    }
                    else if ((tmp_ucb_value = (*nodes_)[i].getUCBValue(parent_trial_count)) > max_ucb_value) {
                        if ((go_board_.isEye((*nodes_)[i].pos_, color_) == false || go_board_.isNick((*nodes_)[i].pos_, color_) == true)
                                && go_board_.isLegalMove((*nodes_)[i].pos_, color_) == true) {
                            max_ucb_value = tmp_ucb_value;
                            max_ucb_value_index = i;
                        }
                        else {
                            (*nodes_)[i].is_legal_move_ = false;
                        }
                    }
                }
                else {
                    range += (range >= kboard_square_size ? 0 : 1);
                }
            }
        }
        else {
            for (long long int i = start_index; i < start_index + kselect_width + sub_range; ++i) {
                if ((*nodes_)[i].is_legal_move_ == true) {
                    if ((*nodes_)[i].trial_count_ == 0) {
                        if (go_board_.isLegalMove((*nodes_)[i].pos_, color_) == true) {
                            return i;
                        }
                        else {
                            (*nodes_)[i].is_legal_move_ = false;
                        }
                    }
                    else if ((*nodes_)[i].trial_count_ > 0) {
                        if ((tmp_ucb_value = (*nodes_)[i].getUCBValue(parent_trial_count)) > max_ucb_value) {
                            max_ucb_value = tmp_ucb_value;
                            max_ucb_value_index = i;
                        }
                    }
                    else if ((tmp_ucb_value = (*nodes_)[i].getUCBValue(parent_trial_count)) > max_ucb_value) {
                        if ((go_board_.isEye((*nodes_)[i].pos_, color_) == false || go_board_.isNick((*nodes_)[i].pos_, color_) == true)
                                && go_board_.isLegalMove((*nodes_)[i].pos_, color_) == true) {
                            max_ucb_value = tmp_ucb_value;
                            max_ucb_value_index = i;
                        }
                        else {
                            (*nodes_)[i].is_legal_move_ = false;
                        }
                    }
                }
                else {
                    if (kselect_width + sub_range < kboard_square_size) {
                        ++sub_range;
                    }

                }
            }
        }

        if (parent_trial_count == ktrial_period) {
            for (int i = start_index + kselect_width; i < range + start_index; ++i) {
                (*nodes_)[i].getUCBValue(parent_trial_count);
            }

            std::sort((*nodes_).begin() + start_index + 1, (*nodes_).begin() + start_index + knode_max_width, UCBNode::descendingOrder);
        }

    }
    return max_ucb_value_index;
}

void UCTThread::doOnePlayOut()
{
    int pos;
    int pass_count = 0;
    int turns = 0;
    std::array<int, kboard_square_size> legal_move;
    std::vector<int> three_empty_move;
    three_empty_move.reserve(20);

    // go_board_.printPatternGoBoard();


    int pibot, pre_position = go_board_.getPrePosition(), move;
    int second_pre_position = go_board_.getSecondPrePosition();


    bool is_set_stone, is_capture;

    for (;;) {
        if (pass_count == 2 || turns > 150) {
            break;
        }

        ///*
        is_capture = false;
        // tori


        is_set_stone = false;
        pibot = abs(normal_distributuion_(engine_));
        pibot = (pibot >= kboard_square_size ? kboard_square_size - 1 : pibot);
        for (int i = pibot; i >= 0; --i) {
            move = make_pattern.transition_count_move_[second_pre_position][pre_position][i].second;
            if ((go_board_.isEye(move, color_) == false || go_board_.isNick(move, color_) == true)
                    && go_board_.isLegalMove(move, color_) == true) {
                go_board_.setStone(move, color_);
                second_pre_position = pre_position;
                pre_position = move;
                color_ = (color_ == Black ? White : Black);
                ++turns;
                pass_count = 0;
                is_set_stone = true;
                break;
            }
        }
        if (is_set_stone) {
            continue;
        }



        for (int i = pibot + 1; i < kboard_square_size; ++i) {
            move = make_pattern.transition_count_move_[second_pre_position][pre_position][i].second;
            if ((go_board_.isEye(move, color_) == false || go_board_.isNick(move, color_) == true)
                    && go_board_.isLegalMove(move, color_) == true) {
                go_board_.setStone(move, color_);
                second_pre_position = pre_position;
                pre_position = move;
                color_ = (color_ == Black ? White : Black);
                ++turns;
                pass_count = 0;
                is_set_stone = true;
                break;
            }
        }
        if (is_set_stone == false) {
            color_ = (color_ == Black ? White : Black);
            ++turns;
            ++pass_count;
        }
    }

    return;

    long long int random_table_index;
    int select_move;
    for (;;) {
        if (pass_count == 2 || turns > 150) {
            break;
        }
        is_capture = false;
        // tori


        if (color_ == Black) {
            for (int i = 0; i < kboard_square_size; ++i) {
                if (go_board_.isEye(board9x9_index[i], White) &&  go_board_.isLegalMove(board9x9_index[i], Black)) {
                    go_board_.setStone(board9x9_index[i], Black);
                    color_ = White;
                    ++turns;
                    pass_count = 0;
                    is_capture = true;
                    break;
                }
            }
        }
        else {
            for (int i = 0; i < kboard_square_size; ++i) {
                if (go_board_.isEye(board9x9_index[i], Black) &&  go_board_.isLegalMove(board9x9_index[i], White)) {
                    go_board_.setStone(board9x9_index[i], White);
                    color_ = Black;
                    ++turns;
                    pass_count = 0;
                    is_capture = true;
                    break;
                }
            }
        }

        if (is_capture) {
            continue;
        }

        is_set_stone = false;
        pibot = pibot_distribution_(engine_);
        random_table_index = random_table_distribution_(engine_);

        for (int i = pibot; i >= 0; --i) {
            select_move = make_pattern.random_table_[random_table_index][i];
            if ((go_board_.isNick(select_move, color_) == true || go_board_.isEye(select_move, color_) == false)
                    && go_board_.isLegalMove(select_move, color_) == true) {
                go_board_.setStone(select_move, color_);
                color_ = (color_ == Black ? White : Black);
                ++turns;
                pass_count = 0;
                is_set_stone = true;
                break;
            }
        }
        if (is_set_stone) {
            continue;
        }
        for (int i = pibot + 1; i < kboard_square_size; ++i) {
            select_move = make_pattern.random_table_[random_table_index][i];
            if ((go_board_.isNick(select_move, color_) == true || go_board_.isEye(select_move, color_) == false)
                    && go_board_.isLegalMove(select_move, color_) == true) {
                go_board_.setStone(select_move, color_);
                color_ = (color_ == Black ? White : Black);
                ++turns;
                pass_count = 0;
                is_set_stone = true;
                break;
            }
        }
        if (is_set_stone) {
            continue;
        }

        ++pass_count;
        color_ = (color_ == Black ? White : Black);
        ++turns;

    }
}

void UCTThread::updateNode(int node_index, int result)
{
    --depth_;
    if (depth_ < -1 || node_index < 0) {
        return;
    }


    if (node_index == 0) {
        ++(*nodes_)[0].trial_count_;
        return;
    }
    (*nodes_)[node_index].win_count_ += result;
     ++(*nodes_)[node_index].trial_count_;

    updateNode((*nodes_)[node_index].parent_index_, (result ? 0 : 1));
}

void UCTThread::makeChildNode(int node_index, int color)
{
    go_board_.setPatternGoBoard();
    std::array<std::pair<int, int>, kboard_square_size> pattern_count = make_pattern.getPatternCountBoard(go_board_, color);
    int start_index =  use_node_max_index_;
    // int pass_index = start_index + kboard_square_size;
    use_node_max_index_ += (kboard_square_size + 2);


    for (int i = 1; i <= kboard_square_size; ++i) {
        (*nodes_)[i + start_index].win_count_ = 0; //pattern_count[i].first;
        (*nodes_)[i + start_index].pos_ = pattern_count[i - 1].second;
        (*nodes_)[i + start_index].root_pos_ = (*nodes_)[node_index].root_pos_;
        (*nodes_)[i + start_index].parent_index_ = node_index;
        (*nodes_)[i + start_index].trial_count_ = 0;
        (*nodes_)[i + start_index].child_index_ = 0;
        (*nodes_)[i + start_index].depth_ = (*nodes_)[node_index].depth_ + 1;
    }
    (*nodes_)[start_index].win_count_ = 0;
    (*nodes_)[start_index].pos_ = 0;
    (*nodes_)[start_index].parent_index_ = node_index;
    (*nodes_)[start_index].trial_count_ = 0;
    (*nodes_)[start_index].child_index_ = 0;
    (*nodes_)[start_index].depth_ = (*nodes_)[node_index].depth_ + 1;
    (*nodes_)[start_index].root_pos_ = (*nodes_)[node_index].root_pos_;
    (*nodes_)[node_index].child_index_ = start_index;
}
